/*	Copyright (c) 1988 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

/*
 * Copyright (c) 1997, by Sun Microsystems, Inc.
 * All rights reserved.
 */

#ifndef _REGEXP0_H
#define	_REGEXP0_H

/*#pragma ident	"@(#)regexp0.h	1.14	97/08/26 SMI"	-- SVr4.0 1.9	*/

#include <string.h>

#ifdef	__cplusplus
extern "C" {
#endif

/* CDS addition:
   IGNORECASE has to be defined (either as a static int or a #def')
*/

#ifndef _REGEXP_H		/* Common part of regexp's */
#define	CBRA	 2	/* 002 */	/* \(  */
#define	CCHR	 4	/* 004 */
#define	CDOT	 8	/* 010 */	/*  .  */
#define	CCL	12	/* 014 */	/* [ ] */
#define	CXCL	16	/* 020 */	/* [ ] */
#define	CDOL	20	/* 024 */	/*  $  */
#define	CCEOF	22	/* 026 */
#define	CKET	24	/* 030 */	/* \)  */
#define	ICHR 	28	/* 034 */	/* Aa  */
#define	CBACK	36	/* 044 */	/* \n  */
#define	NCCL	40	/* 050 */	/* [^] */
#define CLT	32	/* 040 */	/* \<  */
#define CGT	34	/* 042 */	/* \>  */

#define	STAR	01
#define	RNGE	03

#define	NBRA	9

#define	PLACE(c)	ep[c >> 3] |= bittab[c & 07]
#define	ISTHERE(c)	(ep[c >> 3] & bittab[c & 07])
#define	ecmp(s1, s2, n)	(strncmp(s1, s2, n) == 0)
#define isAletter(c)	(((c|040)>='a') && ((c|040)<='z'))
#define isAdigit(c)	((c>='0') && (c<='9'))
#define isAword(c)	(isAletter(c) || isAdigit(c) || (c=='_'))

static char	*braslist[NBRA];
static char	*braelist[NBRA];
static int	sed, nbra;
static char	*loc1, *loc2, *locs;
static int	nodelim;
static int	inaword;

static int	circf;
static int	low;
static int	size;

static unsigned char	bittab[] = { 1, 2, 4, 8, 16, 32, 64, 128 };

#ifdef	__STDC__
int advance(const char *lp, const char *ep);
static void getrnge(const char *str);
#else
int advance();
static void getrnge();
#endif
#endif	/* _REGEXP_H */		/* Common part of regexp's */

char *
#ifdef	__STDC__
compile(char *instring, char *ep, const char *endbuf, int seof)
#else
compile(instring, ep, endbuf, seof)
register char *ep;
char *instring, *endbuf;
int seof;
#endif
{
	INIT	/* Dependent declarations and initializations */
	int c;
	int eof = seof;
	char *lastep;
	int cclcnt;
	char bracket[NBRA], *bracketp;
	int closed;
	int neg;
	int lc;
	int i, cflg;
	int iflag; /* used for non-ascii characters in brackets */

	lastep = NULL;
	if ((c = GETC()) == eof || c == '\n') {
		if (c == '\n') {
			UNGETC(c);
			nodelim = 1;
		}
		if (*ep == 0 && !sed)
			ERROR(41);
		RETURN(ep);
	}
	bracketp = bracket;
	circf = closed = nbra = 0;
	if (c == '^')
		circf++;
	else
		UNGETC(c);
	while (1) {
		if (ep >= endbuf)
			ERROR(50);
		c = GETC();
		if (c != '*' && ((c != '\\') || (PEEKC() != '{'/*}*/)))
			lastep = ep;
		if (c == eof) {
			*ep++ = CCEOF;
			if (bracketp != bracket)
				ERROR(42);
			RETURN(ep);
		}
		switch (c) {

		case '.':
			*ep++ = CDOT;
			continue;

		case '\n':
			if (!sed) {
				UNGETC(c);
				*ep++ = CCEOF;
				nodelim = 1;
				if (bracketp != bracket)
					ERROR(42);
				RETURN(ep);
			} else ERROR(36);
		case '*':
			if (lastep == NULL || *lastep == CBRA ||
			    *lastep == CKET)
				goto defchar;
			*lastep |= STAR;
			continue;

		case '$':
			if (PEEKC() != eof && PEEKC() != '\n')
				goto defchar;
			*ep++ = CDOL;
			continue;

		case '[':
			if (&ep[17] >= endbuf)
				ERROR(50);

			*ep++ = CCL;
			lc = 0;
			for (i = 0; i < 16; i++)
				ep[i] = 0;

			neg = 0;
			if ((c = GETC()) == '^') {
				neg = 1;
				c = GETC();
			}
			iflag = 1;
			do {
				c &= 0377;
				if (c == '\0' || c == '\n')
					ERROR(49);
				if ((c & 0200) && iflag) {
					iflag = 0;
					if (&ep[32] >= endbuf)
						ERROR(50);
					ep[-1] = CXCL;
					for (i = 16; i < 32; i++)
						ep[i] = 0;
				}
				if (c == '-' && lc != 0) {
					if ((c = GETC()) == ']') {
						PLACE('-');
						break;
					}
					if ((c & 0200) && iflag) {
						iflag = 0;
						if (&ep[32] >= endbuf)
							ERROR(50);
						ep[-1] = CXCL;
						for (i = 16; i < 32; i++)
							ep[i] = 0;
					}
					if (IGNORECASE && isAletter(c)) {
						c |= 040;
						iflag |= 040;
					}
					while (lc < c) {
						PLACE(lc);
						if (iflag&040) {
						    lc ^= 040;	/* low<->up */
						    PLACE(lc);
						    lc ^= 040;
						}
						lc++;
					}
					iflag &= ~040;
				}
				lc = c;
				PLACE(c);
				if (IGNORECASE && isAletter(c)) {
				    c ^= 040;
				    PLACE(c);
				    lc = c | 040;	/* Keep lowercase */
				}
			} while ((c = GETC()) != ']');

			if (iflag)
				iflag = 16;
			else
				iflag = 32;

			if (neg) {
				if (iflag == 32) {
					for (cclcnt = 0; cclcnt < iflag;
					    cclcnt++)
						ep[cclcnt] ^= 0377;
					ep[0] &= 0376;
				} else {
					ep[-1] = NCCL;
					/* make nulls match so test fails */
					ep[0] |= 01;
				}
			}

			ep += iflag;

			continue;

		case '\\':
			switch (c = GETC()) {

			case '(':
				if (nbra >= NBRA)
					ERROR(43);
				*bracketp++ = (char)nbra;
				*ep++ = CBRA;
				*ep++ = (char)nbra++;
				continue;

			case ')':
				if (bracketp <= bracket)
					ERROR(42);
				*ep++ = CKET;
				*ep++ = *--bracketp;
				closed++;
				continue;

			case '<':
				*ep++ = CLT;
				continue;
			case '>':
				*ep++ = CGT;
				continue;

			case '{': /*}*/
				if (lastep == NULL)
					goto defchar;
				if (&ep[2] >= endbuf)
					ERROR(50);
				*lastep |= RNGE;
				cflg = 0;
			nlim:
				c = GETC();
				i = 0;
				do {
					if ('0' <= c && c <= '9')
						i = 10 * i + c - '0';
					else
						ERROR(16);
				} while (((c = GETC()) != '\\') && (c != ','));
				if (i >= 255)
					ERROR(11);
				*ep++ = (char)i;
				if (c == ',') {
					if (cflg++)
						ERROR(44);
					if ((c = GETC()) == '\\')
						*ep++ = (char)255;
					else {
						UNGETC(c);
						goto nlim;
						/* get 2'nd number */
					}
				}
				if (GETC() != /*{*/'}')
					ERROR(45);
				if (!cflg)	/* one number */
					*ep++ = (char)i;
				else if ((ep[-1] & 0377) < (ep[-2] & 0377))
					ERROR(46);
				continue;

			case '\n':
				ERROR(36);

			case 'n':
				c = '\n';
				goto defchar;

			default:
				if (c >= '1' && c <= '9') {
					if ((c -= '1') >= closed)
						ERROR(25);
					*ep++ = CBACK;
					*ep++ = (char)c;
					continue;
				}
			}
	/* Drop through to default to use \ to turn off special chars */

		defchar:
		default:
			lastep = ep;
			if (IGNORECASE && isAletter(c)) { 
				c |= 040;	/* Keep lowercase */
				*ep++ = ICHR;
				*ep++ = (char)c;

			}
			else {
				*ep++ = CCHR;
				*ep++ = (char)c;
			}
		}
	}
}

#ifndef _REGEXP_H		/* Common part of regexp's */
#ifdef	__STDC__
int
step(const char *p1, const char *p2)
#else
int
step(p1, p2)
register char *p1, *p2;
#endif
{
	char c;

	inaword = 0;
	if (circf) {
		loc1 = (char *) p1;
		return (advance(p1, p2));
	}
	/* fast check for first character */
	if (*p2 == CCHR) {
		c = p2[1];
		do {
			if (*p1 != c)
				continue;
			if (advance(p1, p2)) {
				loc1 = (char *) p1;
				return (1);
			}
		} while (*p1++);
		return (0);
	}
#if 0		/* Original algorithm -- problems with CLT/CGT! */
		/* regular algorithm */
	do {
		if (advance(p1, p2)) {
			loc1 = (char *) p1;
			return (1);
		}
	} while (*p1++);
#else
	while (*p1) {
		if (advance(p1, p2)) {
			loc1 = (char *) p1;
			return (1);
		}
		if ((*p2 == CLT) && isAword(*p1)) {	/* Skip Word */
		    inaword = 0;
		    while (isAword(*p1)) p1++;
		}
		else p1++;
	}
#endif
	return (0);
}

int
#ifdef	__STDC__
advance(const char *lp, const char *ep)
#else
advance(lp, ep)
register char *lp, *ep;
#endif
{
#ifdef	__STDC__
	const char *curlp;
#else
	register char *curlp;
#endif
	int c, cc;
	char *bbeg;
	register char neg;
	size_t ct;

	while (1) {
		neg = 0;
#if 0		/* Could accept only alphanumeric words ? */
		if ((inaword&1) && *ep != CGT) {
			if (isAword(*lp)) ;
			else return(0);
		}
#endif

		switch (*ep++) {

		case CCHR:
			if (*ep++ == *lp++)
				continue;
			return (0);
			/*FALLTHRU*/

		case ICHR:
			c = *lp++;
			if (isAletter(c)) c |= 040;
			if (*ep++ == c)
				continue;
			return (0);
			/*FALLTHRU*/

		case CDOT:
			if (*lp++)
				continue;
			return (0);
			/*FALLTHRU*/

		case CDOL:
			if (*lp == 0)
				continue;
			return (0);
			/*FALLTHRU*/

		case CCEOF:
			loc2 = (char *) lp;
			return (1);
			/*FALLTHRU*/

		case CXCL:
			c = (unsigned char)*lp++;
			if (ISTHERE(c)) {
				ep += 32;
				continue;
			}
			return (0);
			/*FALLTHRU*/

		case NCCL:
			neg = 1;
			/*FALLTHRU*/

		case CCL:
			c = *lp++;
			if (((c & 0200) == 0 && ISTHERE(c)) ^ neg) {
				ep += 16;
				continue;
			}
			return (0);
			/*FALLTHRU*/

		case CBRA:
			braslist[(*ep++)&0xff] = (char *) lp;
			continue;
			/*FALLTHRU*/

		case CKET:
			braelist[(*ep++)&0xff] = (char *) lp;
			continue;
			/*FALLTHRU*/

		case ICHR | RNGE:
			cc = *ep++;
			getrnge(ep);
			while (low--) {
				c = *lp++;
				if (isAletter(c)) c |= 040;
				if (c != cc)
					return (0);
			}
			curlp = lp;
			while (size--) {
				c = *lp++;
				if (isAletter(c)) c |= 040;
				if (c != cc)
					break;
			}
			if (size < 0)
				lp++;
			ep += 2;
			goto star;
			/*FALLTHRU*/

		case CCHR | RNGE:
			cc = *ep++;
			getrnge(ep);
			while (low--)
				if (*lp++ != cc)
					return (0);
			curlp = lp;
			while (size--)
				if (*lp++ != cc)
					break;
			if (size < 0)
				lp++;
			ep += 2;
			goto star;
			/*FALLTHRU*/

		case CDOT | RNGE:
			getrnge(ep);
			while (low--)
				if (*lp++ == '\0')
					return (0);
			curlp = lp;
			while (size--)
				if (*lp++ == '\0')
					break;
			if (size < 0)
				lp++;
			ep += 2;
			goto star;
			/*FALLTHRU*/

		case CXCL | RNGE:
			getrnge(ep + 32);
			while (low--) {
				c = (unsigned char)*lp++;
				if (!ISTHERE(c))
					return (0);
			}
			curlp = lp;
			while (size--) {
				c = (unsigned char)*lp++;
				if (!ISTHERE(c))
					break;
			}
			if (size < 0)
				lp++;
			ep += 34;		/* 32 + 2 */
			goto star;
			/*FALLTHRU*/

		case NCCL | RNGE:
			neg = 1;
			/*FALLTHRU*/

		case CCL | RNGE:
			getrnge(ep + 16);
			while (low--) {
				c = *lp++;
				if (((c & 0200) || !ISTHERE(c)) ^ neg)
					return (0);
			}
			curlp = lp;
			while (size--) {
				c = *lp++;
				if (((c & 0200) || !ISTHERE(c)) ^ neg)
					break;
			}
			if (size < 0)
				lp++;
			ep += 18; 		/* 16 + 2 */
			goto star;
			/*FALLTHRU*/

		case CBACK:
			bbeg = braslist[(*ep)&0xff];
			ct = braelist[(*ep++)&0xff] - bbeg;

			if (ecmp(bbeg, lp, ct)) {
				lp += ct;
				continue;
			}
			return (0);
			/*FALLTHRU*/

		case CLT:
			if (inaword&1) 
				return(0);
			if ((inaword&2) && isAword(lp[-1])) 
				return(0);
			if (isAword(*lp)) inaword = 1;
			else return(0);
			continue;
			/*FALLTHRU*/

		case CGT:
			if (isAword(*lp)) return(0);
			if (inaword&1) inaword = 2;
			else return(0);
			continue;
			/*FALLTHRU*/

		case CBACK | STAR:
			bbeg = braslist[(*ep)&0xff];
			ct = braelist[(*ep++)&0xff] - bbeg;
			curlp = lp;
			while (ecmp(bbeg, lp, ct))
				lp += ct;

			while (lp >= curlp) {
				if (advance(lp, ep))
					return (1);
				lp -= ct;
			}
			return (0);
			/*FALLTHRU*/

		case CDOT | STAR:
			curlp = lp;
			while (*lp++);
			goto star;
			/*FALLTHRU*/

		case ICHR | STAR:
			curlp = lp;
			cc = *ep++;
			while (1) {
				c = *lp++;
				if (isAletter(c)) c |= 040;
				if (c != cc) break;
			}
			goto star;
			/*FALLTHRU*/

		case CCHR | STAR:
			curlp = lp;
			while (*lp++ == *ep);
			ep++;
			goto star;
			/*FALLTHRU*/

		case CXCL | STAR:
			curlp = lp;
			do {
				c = (unsigned char)*lp++;
			} while (ISTHERE(c));
			ep += 32;
			goto star;
			/*FALLTHRU*/

		case NCCL | STAR:
			neg = 1;
			/*FALLTHRU*/

		case CCL | STAR:
			curlp = lp;
			do {
				c = *lp++;
			} while (((c & 0200) == 0 && ISTHERE(c)) ^ neg);
			ep += 16;
			goto star;
			/*FALLTHRU*/

		star:
			do {
				if (--lp == locs)
					break;
				if (advance(lp, ep))
					return (1);
			} while (lp > curlp);
			return (0);

		}
	}
}

static void
#ifdef	__STDC__
getrnge(const char *str)
#else
getrnge(str)
register char *str;
#endif
{
	low = *str++ & 0377;
	size = ((*str & 0377) == 255)? 20000: (*str &0377) - low;
}

#define _REGEXP_H
#endif	/* _REGEXP_H */		/* Common part of regexp's */

#ifdef	__cplusplus
}
#endif

#endif	/* _REGEXP0_H */
